#  SPDX-License-Identifier: Apache-2.0
"""
subarulink - A Python Package for interacting with Subaru Starlink Remote Services API.

controller.py - provides managed connection to Subaru API

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""
import asyncio
from datetime import datetime
import json
import logging
import pprint
import time

from subarulink.connection import Connection
import subarulink.const as sc
from subarulink.exceptions import InvalidPIN, SubaruException

_LOGGER = logging.getLogger(__name__)


class Controller:
    """Controller for connections to Subaru Starlink API."""

    def __init__(
        self, websession, username, password, device_id, pin, device_name, update_interval=7200, fetch_interval=300,
    ):
        """Initialize controller.

        Args:
            websession (aiohttp.ClientSession): Session
            username (Text): Username
            password (Text): Password
            device_id (Text): Alphanumeric designator that Subaru API uses to determine if a device is authorized to send remote requests
            pin (Text): 4 digit pin number string required to submit Subaru Remote requests
            device_name (Text): Human friendly name that is associated with device_id (shows on mysubaru.com profile "devices")
            update_interval (int, optional): Seconds between requests for vehicle send update
            fetch_interval (int, optional): Seconds between fetches of Subaru's cached vehicle information

        """
        self._connection = Connection(websession, username, password, device_id, device_name)
        self._update_interval = update_interval
        self._fetch_interval = fetch_interval
        self._car_data = {}
        self._update = {}
        self._pin = pin
        self._vin_id_map = {}
        self._vin_name_map = {}
        self._api_gen = {}
        self._lock = {}
        self._hasEV = {}
        self._hasRemote = {}
        self._hasRES = {}
        self._hasSafety = {}
        self._controller_lock = asyncio.Lock()
        self._last_update_time = {}
        self._last_fetch_time = {}
        self._cars = []

    async def connect(self, test_login=False) -> bool:
        """
        Connect to Subaru Remote Services API.

        Args:
            test_login (Bool, optional): Only check for authorization

        """
        if test_login:
            response = await self._connection.connect()
            if response:
                return True
            return False

        _LOGGER.debug("Connecting controller to Subaru Remote Services")
        cars = await self._connection.connect()
        if cars is None:
            raise SubaruException("Connection to Subaru API failed")

        for car in cars:
            vin = car["vin"].upper()
            self._cars.append(vin)
            self._vin_name_map[vin] = car["display_name"]
            self._vin_id_map[vin] = car["id"]
            self._api_gen[vin] = car["api_gen"]
            self._hasEV[vin] = car["hasEV"]
            self._hasRES[vin] = car["hasRES"]
            self._hasRemote[vin] = car["hasRemote"]
            self._hasSafety[vin] = car["hasSafety"]
            self._lock[vin] = asyncio.Lock()
            self._last_update_time[vin] = 0
            self._last_fetch_time[vin] = 0
            self._car_data[vin] = {"status": {}}
            self._update[vin] = True
        _LOGGER.debug("Subaru Remote Services Ready!")
        return True

    def get_vehicles(self):
        """Return list of VINs available to user on Subaru Remote Services API."""
        return self._cars

    def get_ev_status(self, vin):
        """Get if EV."""
        _LOGGER.debug("Getting EV Status %s:%s", vin, str(self._hasEV.get(vin)))
        return self._hasEV.get(vin.upper())

    def get_remote_status(self, vin):
        """Get if remote services available."""
        _LOGGER.debug("Getting remote Status %s:%s", vin, str(self._hasRemote.get(vin)))
        return self._hasRemote.get(vin.upper())

    def get_res_status(self, vin):
        """Get if remote engine start is available."""
        _LOGGER.debug("Getting RES Status %s:%s", vin, str(self._hasRES.get(vin)))
        return self._hasRES.get(vin.upper())

    def get_safety_status(self, vin):
        """Get if safety plus subscription is active."""
        _LOGGER.debug("Getting Safety Plus Status %s:%s", vin, str(self._hasSafety.get(vin)))
        return self._hasSafety.get(vin.upper())

    def get_api_gen(self, vin):
        """Get API version (g1 or g2) for vehicle."""
        return self._api_gen.get(vin.upper())

    def vin_to_name(self, vin):
        """Return display name for a given VIN."""
        return self._vin_name_map.get(vin.upper())

    async def get_data(self, vin):
        """Get locally cached vehicle data.  Fetch if not present."""
        vin = vin.upper()
        if len(self._car_data[vin]["status"]) == 0:
            await self.fetch(vin)
        return self._car_data[vin]

    async def get_climate_settings(self, vin):
        """Fetch saved climate control settings."""
        vin = vin.upper()
        if self._hasRES[vin] or self._hasEV[vin]:
            await self._connection.validate_session(vin)
            js_resp = await self._get("service/g2/remoteEngineStart/fetch.json")
            _LOGGER.debug(js_resp)
            if js_resp["success"]:
                self._car_data[vin]["climate"] = json.loads(js_resp["data"])
                return True
            else:
                _LOGGER.error("Failed to fetch saved climate settings: %s", js_resp["errorCode"])

    async def save_climate_settings(self, vin, form_data):
        """Fetch saved climate control settings."""
        vin = vin.upper()
        if self._hasRES[vin] or self._hasEV[vin]:
            if self._validate_remote_start_params(vin, form_data):
                js_resp = await self._post("/service/g2/remoteEngineStart/save.json", json=form_data)
                _LOGGER.debug(js_resp)
                if js_resp["success"]:
                    self._car_data[vin]["climate"] = js_resp["data"]
                    _LOGGER.info("Climate control settings saved.")
                    return True
            else:
                _LOGGER.error("Failed to fetch saved climate settings: %s", js_resp["errorCode"])

    async def fetch(self, vin, force=False):
        """Fetch latest data from Subaru.  Does not invoke a remote request."""
        vin = vin.upper()
        cur_time = time.time()
        async with self._controller_lock:
            last_fetch = self._last_fetch_time[vin]
            if force or cur_time - last_fetch > self._fetch_interval:
                await self._fetch_status(vin)
                self._last_fetch_time[vin] = cur_time

    async def update(self, vin, force=False):
        """Request Subaru send remote command to update vehicle data."""
        vin = vin.upper()
        cur_time = time.time()
        async with self._controller_lock:
            last_update = self._last_update_time[vin]
            if force or cur_time - last_update > self._update_interval:
                result = await self._locate(vin)
                await self._fetch_status(vin)
                self._last_update_time[vin] = cur_time
                return result

    def get_update_interval(self):
        """Get current update interval."""
        return self._update_interval

    def set_update_interval(self, value):
        """Set new update interval."""
        old_interval = self._update_interval
        if value >= 300:
            self._update_interval = value
            _LOGGER.debug("Update interval changed from %s to %s", old_interval, value)
        else:
            _LOGGER.error("Invalid update interval %s. Keeping old value: %s", value, old_interval)

    def get_fetch_interval(self):
        """Get current fetch interval."""
        return self._fetch_interval

    def set_fetch_interval(self, value):
        """Set new fetch interval."""
        old_interval = self._fetch_interval
        if value >= 60:
            self._fetch_interval = value
            _LOGGER.debug("Fetch interval changed from %s to %s", old_interval, value)
        else:
            _LOGGER.error("Invalid fetch interval %s. Keeping old value: %s", value, old_interval)

    def get_last_update_time(self, vin):
        """Get last time update() remote command was used."""
        return self._last_update_time[vin.upper()]

    async def charge_start(self, vin):
        """Start EV charging."""
        success, _ = await self._remote_command(vin.upper(), "phevChargeNow")
        return success

    async def lock(self, vin):
        """Send lock command."""
        form_data = {"forceKeyInCar": False}
        success, _ = await self._actuate(vin.upper(), "lock", data=form_data)
        return success

    async def unlock(self, vin, only_driver=True):
        """Send unlock command."""
        door = sc.ALL_DOORS
        if only_driver:
            door = sc.DRIVERS_DOOR
        form_data = {sc.WHICH_DOOR: door}
        success, _ = await self._actuate(vin.upper(), "unlock", data=form_data)
        return success

    async def lights(self, vin):
        """Send lights command."""
        success, _ = await self._actuate(vin.upper(), "lightsOnly")
        return success

    async def horn(self, vin):
        """Send horn command."""
        success, _ = await self._actuate(vin.upper(), "hornLights")
        return success

    async def remote_stop(self, vin):
        """Send remote stop command."""
        success, _ = await self._actuate(vin.upper(), "engineStop")
        return success

    async def remote_start(self, vin, form_data=None):
        """Send remote start command."""
        vin = vin.upper()
        if self._hasRES[vin] or self._hasEV[vin]:
            if form_data:
                if self._validate_remote_start_params(vin, form_data):
                    climate_settings = form_data
                else:
                    raise SubaruException("Error with climate settings")
            else:
                await self.get_climate_settings(vin)
                climate_settings = self._car_data[vin]["climate"]
            if climate_settings:
                success, _ = await self._actuate(vin, "engineStart", data=climate_settings)
                return success
            else:
                raise SubaruException("Error with climate settings")
        else:
            raise SubaruException("Remote Start not supported for this vehicle")

    def get_updates(self, vin):
        """Get updates dictionary.

        Parameters
        ----------
        vin : string
            VIN for the vehicle.

        Returns
        -------
        bool or None
            If vin exists, a bool indicating whether updates should be
            processed. Othewise, returns None.

        """
        vin = vin.upper()
        if vin in self._update:
            return self._update[vin]
        return None

    def set_updates(self, vin, setting):
        """Change update setting for vehicle."""
        self._update[vin.upper()] = setting

    async def _get(self, cmd, params=None, data=None, json=None):
        return await self._connection.get("/%s" % cmd, params, data, json)

    async def _post(self, cmd, params=None, data=None, json=None):
        return await self._connection.post("/%s" % cmd, params, data, json)

    async def _remote_query(self, vin, cmd, data=None):
        await self._connection.validate_session(vin)
        api_gen = self._api_gen[vin]
        async with self._lock[vin]:
            js_resp = await self._get("service/%s/%s/execute.json" % (api_gen, cmd), json=data)
            _LOGGER.debug(pprint.pformat(js_resp))
            if js_resp["success"]:
                return js_resp
            raise SubaruException("Remote query failed. Response: %s " % js_resp)

    async def _remote_command(self, vin, cmd, data=None, poll_url="/service/api_gen/remoteService/status.json"):
        await self._connection.validate_session(vin)
        api_gen = self._api_gen[vin]
        form_data = {"pin": self._pin}
        if data:
            form_data.update(data)
        req_id = ""
        async with self._lock[vin]:
            js_resp = await self._post("service/%s/%s/execute.json" % (api_gen, cmd), json=form_data)
            _LOGGER.debug(pprint.pformat(js_resp))
            if js_resp["success"]:
                req_id = js_resp["data"][sc.SERVICE_REQ_ID]
                return await self._wait_request_status(req_id, api_gen, poll_url)
            if js_resp["errorCode"] == "InvalidCredentials":
                raise InvalidPIN(js_resp["data"]["errorDescription"])
            raise SubaruException("Remote command failed.  Response: %s " % js_resp)

    async def _actuate(self, vin, cmd, data=None):
        form_data = {"delay": 0, "vin": vin}
        if data:
            form_data.update(data)
        if self.get_remote_status(vin):
            return await self._remote_command(vin, cmd, data=form_data)
        else:
            raise SubaruException("Command requires REMOTE subscription.")

    async def _get_vehicle_status(self, vin):
        await self._connection.validate_session(vin)
        js_resp = await self._get("vehicleStatus.json")
        _LOGGER.debug(pprint.pformat(js_resp))
        return js_resp

    async def _fetch_status(self, vin):
        if self.get_safety_status(vin):
            _LOGGER.debug("Fetching vehicle status from Subaru")
            js_resp = await self._get_vehicle_status(vin)
            if js_resp.get("success") and js_resp.get("data"):
                data = js_resp["data"]
                status = {}
                status[sc.ODOMETER] = data.get(sc.VS_ODOMETER)
                status[sc.TIMESTAMP] = data.get(sc.VS_TIMESTAMP) / 1000
                status[sc.HEADING] = data.get(sc.VS_HEADING)
                status[sc.AVG_FUEL_CONSUMPTION] = data.get(sc.VS_AVG_FUEL_CONSUMPTION)
                status[sc.DIST_TO_EMPTY] = data.get(sc.VS_DIST_TO_EMPTY)
                status[sc.VEHICLE_STATE] = data.get(sc.VS_VEHICLE_STATE)
                status[sc.TIRE_PRESSURE_FL] = int(data.get(sc.VS_TIRE_PRESSURE_FL) or 0)
                status[sc.TIRE_PRESSURE_FR] = int(data.get(sc.VS_TIRE_PRESSURE_FR) or 0)
                status[sc.TIRE_PRESSURE_RL] = int(data.get(sc.VS_TIRE_PRESSURE_RL) or 0)
                status[sc.TIRE_PRESSURE_RR] = int(data.get(sc.VS_TIRE_PRESSURE_RR) or 0)
                if data.get(sc.VS_LONGITUDE) != sc.BAD_LONGITUDE:
                    status[sc.LONGITUDE] = data.get(sc.VS_LONGITUDE)
                if data.get(sc.VS_LATITUDE) != sc.BAD_LATITUDE:
                    status[sc.LATITUDE] = data.get(sc.VS_LATITUDE)
                if data.get(sc.VS_HEADING):
                    status[sc.HEADING] = data.get(sc.VS_HEADING)

                self._car_data[vin]["status"].update(status)
            else:
                raise SubaruException("Error fetching vehicle status %s" % pprint.pformat(js_resp))
        else:
            raise SubaruException("Safety Plus subscription required for this vehicle")

        # Additional Data (Security Plus Required)
        if self.get_remote_status(vin):
            js_resp = await self._remote_query(vin, "condition")
            if js_resp.get("success") and js_resp.get("data"):
                status = await self._cleanup_condition(js_resp, vin)
                self._car_data[vin]["status"].update(status)

    async def _cleanup_condition(self, js_resp, vin):
        data = {}
        try:
            # Annoying key/value pair format [{"key": key, "value": value}, ...]
            data = {i["key"]: i["value"] for i in js_resp["data"]["result"]["vehicleStatus"]}
            data[sc.TIMESTAMP] = datetime.strptime(
                js_resp["data"]["result"]["lastUpdatedTime"], sc.TIMESTAMP_FMT
            ).timestamp()
            data[sc.POSITION_TIMESTAMP] = datetime.strptime(
                data[sc.POSITION_TIMESTAMP], sc.POSITION_TIMESTAMP_FMT
            ).timestamp()

            # Discard these values since vehicleStatus.json is always more reliable
            data.pop(sc.ODOMETER)
            data.pop(sc.AVG_FUEL_CONSUMPTION)
            data.pop(sc.DIST_TO_EMPTY)
            data.pop(sc.TIRE_PRESSURE_FL)
            data.pop(sc.TIRE_PRESSURE_FR)
            data.pop(sc.TIRE_PRESSURE_RL)
            data.pop(sc.TIRE_PRESSURE_RR)

            if self.get_ev_status(vin):
                if int(data.get(sc.EV_DISTANCE_TO_EMPTY) or 0) > 20:
                    # This value is incorrectly high immediately after car shutdown
                    data.pop(sc.EV_DISTANCE_TO_EMPTY)
                if (
                    int(data.get(sc.EV_TIME_TO_FULLY_CHARGED) or sc.BAD_EV_TIME_TO_FULLY_CHARGED)
                    == sc.BAD_EV_TIME_TO_FULLY_CHARGED
                ):
                    # Value is None or known erroneous number
                    data[sc.EV_TIME_TO_FULLY_CHARGED] = 0
                # Value is correct unless it is None
                data[sc.EV_DISTANCE_TO_EMPTY] = int(data.get(sc.EV_DISTANCE_TO_EMPTY) or 0)

            if not self._car_data[vin]["status"].get(sc.LATITUDE) or not self._car_data[vin]["status"].get(
                sc.LONGITUDE
            ):
                _LOGGER.warning("Location invalid, requesting vehicle update, this may take up to 20 seconds")
                await self._locate(vin)
        except KeyError:  # Once in a while a 'value' key or some other field is missing
            pass

        return data

    async def _locate(self, vin):
        success, js_resp = await self._remote_command(
            vin, "vehicleStatus", poll_url="/service/api_gen/vehicleStatus/locationStatus.json",
        )
        if success and js_resp.get("success"):
            self._car_data[vin]["status"][sc.LONGITUDE] = js_resp["data"]["result"][sc.LONGITUDE]
            self._car_data[vin]["status"][sc.LATITUDE] = js_resp["data"]["result"][sc.LATITUDE]
            self._car_data[vin]["status"][sc.HEADING] = js_resp["data"]["result"][sc.HEADING]
            return True

    async def _wait_request_status(self, req_id, api_gen, poll_url, attempts=20):
        params = {sc.SERVICE_REQ_ID: req_id}
        attempt = 0
        _LOGGER.debug("Polling for remote service request completion: serviceRequestId=%s", req_id)
        while attempt < attempts:
            try:
                js_resp = await self._connection.get(poll_url.replace("api_gen", api_gen), params=params)
            except SubaruException as ex:
                attempt += 1
                _LOGGER.error("Remote service status poll request returned error %s", ex.message)
                continue
            _LOGGER.debug(pprint.pformat(js_resp))
            if not js_resp["success"]:
                _LOGGER.error("Remote service command returned error: %s", js_resp["errorCode"])
            elif js_resp["data"]["remoteServiceState"] == "finished":
                if js_resp["data"]["success"]:
                    _LOGGER.info("Remote service request completed successfully: %s", req_id)
                    return True, js_resp
                else:
                    _LOGGER.error(
                        "Remote service request completed but failed: %s Error: %s",
                        req_id,
                        js_resp["data"]["errorCode"],
                    )
                    return False, js_resp
            elif js_resp["data"].get("remoteServiceState") == "started":
                _LOGGER.info("Subaru API reports remote service request is in progress: %s", req_id)
            attempt += 1
            await asyncio.sleep(2)
        _LOGGER.error("Remote service request completion message not received")
        return False, None

    def _validate_remote_start_params(self, vin, form_data):
        try:
            temp = int(form_data[sc.TEMP])
            is_valid = True
            if temp > sc.TEMP_MAX or temp < sc.TEMP_MIN:
                is_valid = False
            if form_data[sc.MODE] not in [
                sc.MODE_AUTO,
                sc.MODE_DEFROST,
                sc.MODE_FACE,
                sc.MODE_FEET,
                sc.MODE_FEET_DEFROST,
                sc.MODE_SPLIT,
            ]:
                is_valid = False
            if form_data[sc.HEAT_SEAT_LEFT] not in [
                sc.HEAT_SEAT_OFF,
                sc.HEAT_SEAT_HI,
                sc.HEAT_SEAT_MED,
                sc.HEAT_SEAT_LOW,
            ]:
                is_valid = False
            if form_data[sc.HEAT_SEAT_RIGHT] not in [
                sc.HEAT_SEAT_OFF,
                sc.HEAT_SEAT_HI,
                sc.HEAT_SEAT_MED,
                sc.HEAT_SEAT_LOW,
            ]:
                is_valid = False
            if form_data[sc.REAR_DEFROST] not in [sc.REAR_DEFROST_OFF, sc.REAR_DEFROST_ON]:
                is_valid = False
            if form_data[sc.FAN_SPEED] not in [
                sc.FAN_SPEED_AUTO,
                sc.FAN_SPEED_HI,
                sc.FAN_SPEED_LOW,
                sc.FAN_SPEED_MED,
            ]:
                is_valid = False
            if form_data[sc.RECIRCULATE] not in [sc.RECIRCULATE_OFF, sc.RECIRCULATE_ON]:
                is_valid = False
            if form_data[sc.REAR_AC] not in [sc.REAR_AC_OFF, sc.REAR_AC_ON]:
                is_valid = False

            form_data[sc.RUNTIME] = sc.RUNTIME_DEFAULT
            form_data[sc.CLIMATE] = sc.CLIMATE_DEFAULT
            if self._hasEV[vin]:
                form_data[sc.START_CONFIG] = sc.START_CONFIG_DEFAULT_EV
            elif self._hasRES[vin]:
                form_data[sc.START_CONFIG] = sc.START_CONFIG_DEFAULT_RES
            else:
                raise SubaruException("Vehicle Remote Start not supported.")

            return is_valid

        except KeyError:
            return None



#  SPDX-License-Identifier: Apache-2.0
"""
subarulink - A Python Package for interacting with Subaru Starlink Remote Services API.

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""

import subarulink.const as const
from subarulink.controller import Controller
from subarulink.exceptions import IncompleteCredentials, InvalidPIN, SubaruException

from .__version__ import __version__

__all__ = [
    "Controller",
    "SubaruException",
    "InvalidPIN",
    "IncompleteCredentials",
    "const",
    "__version__",
]



#  SPDX-License-Identifier: Apache-2.0
"""
subarulink - A Python Package for interacting with Subaru Starlink Remote Services API.

exceptions.py - provides exceptions specific to the package

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""


class SubaruException(Exception):
    """Class of Subaru API exceptions."""

    def __init__(self, message, *args, **kwargs):
        """Initialize exceptions for the Subaru Starlink API."""
        self.message = message
        super().__init__(*args, **kwargs)


class InvalidPIN(SubaruException):
    """Class of exceptions for invalid PIN number."""

    def __init__(self, *args, **kwargs):
        # pylint: disable=super-init-not-called
        """Initialize exception."""
        pass


class RetryLimitError(SubaruException):
    """Class of exceptions for hitting retry limits."""

    def __init__(self, *args, **kwargs):
        # pylint: disable=super-init-not-called
        """Initialize exception."""
        pass


class IncompleteCredentials(SubaruException):
    """Class of exceptions for hitting retry limits."""

    def __init__(self, *args, **kwargs):
        # pylint: disable=super-init-not-called
        """Initialize exception."""
        pass



#  SPDX-License-Identifier: Apache-2.0
"""
subarulink - A Python Package for interacting with Subaru Starlink Remote Services API.

const.py - constants used with this package

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""

SERVICE_REQ_ID = "serviceRequestId"

# Remote start constants
TEMP = "climateZoneFrontTemp"
TEMP_MAX = 85
TEMP_MIN = 60

CLIMATE = "climateSettings"
CLIMATE_DEFAULT = "climateSettings"

RUNTIME = "runTimeMinutes"
RUNTIME_DEFAULT = "10"

MODE = "climateZoneFrontAirMode"
MODE_DEFROST = "WINDOW"
MODE_FEET_DEFROST = "FEET_WINDOW"
MODE_FACE = "FACE"
MODE_FEET = "FEET"
MODE_SPLIT = "FEET_FACE_BALANCED"
MODE_AUTO = "AUTO"

HEAT_SEAT_LEFT = "heatedSeatFrontLeft"
HEAT_SEAT_RIGHT = "heatedSeatFrontRight"
HEAT_SEAT_HI = "HIGH_HEAT"
HEAT_SEAT_MED = "MEDIUM_HEAT"
HEAT_SEAT_LOW = "LOW_HEAT"
HEAT_SEAT_OFF = "OFF"

REAR_DEFROST = "heatedRearWindowActive"
REAR_DEFROST_ON = "true"
REAR_DEFROST_OFF = "false"

FAN_SPEED = "climateZoneFrontAirVolume"
FAN_SPEED_LOW = "2"
FAN_SPEED_MED = "4"
FAN_SPEED_HI = "7"
FAN_SPEED_AUTO = "AUTO"

RECIRCULATE = "outerAirCirculation"
RECIRCULATE_OFF = "outsideAir"
RECIRCULATE_ON = "recirculation"

REAR_AC = "airConditionOn"
REAR_AC_ON = "true"
REAR_AC_OFF = "false"

START_CONFIG = "startConfiguration"
START_CONFIG_DEFAULT_EV = "start_Climate_Control_only_allow_key_in_ignition"
START_CONFIG_DEFAULT_RES = "START_ENGINE_ALLOW_KEY_IN_IGNITION"


# Unlock doors constants
WHICH_DOOR = "unlockDoorType"
ALL_DOORS = "ALL_DOORS_CMD"
DRIVERS_DOOR = "FRONT_LEFT_DOOR_CMD"

# Location data constants
HEADING = "heading"
LATITUDE = "latitude"
LONGITUDE = "longitude"
LOCATION_TIME = "locationTimestamp"
SPEED = "speed"
BAD_LATITUDE = 90.0
BAD_LONGITUDE = 180.0

# Vehicle status constants
AVG_FUEL_CONSUMPTION = "AVG_FUEL_CONSUMPTION"
BATTERY_VOLTAGE = "BATTERY_VOLTAGE"
DIST_TO_EMPTY = "DISTANCE_TO_EMPTY_FUEL"
DOOR_BOOT_POSITION = "DOOR_BOOT_POSITION"
DOOR_ENGINE_HOOD_POSITION = "DOOR_ENGINE_HOOD_POSITION"
DOOR_FRONT_LEFT_POSITION = "DOOR_FRONT_LEFT_POSITION"
DOOR_FRONT_RIGHT_POSITION = "DOOR_FRONT_RIGHT_POSITION"
DOOR_REAR_LEFT_POSITION = "DOOR_REAR_LEFT_POSITION"
DOOR_REAR_RIGHT_POSITION = "DOOR_REAR_RIGHT_POSITION"
EV_CHARGER_STATE_TYPE = "EV_CHARGER_STATE_TYPE"
EV_CHARGE_SETTING_AMPERE_TYPE = "EV_CHARGE_SETTING_AMPERE_TYPE"
EV_CHARGE_VOLT_TYPE = "EV_CHARGE_VOLT_TYPE"
EV_DISTANCE_TO_EMPTY = "EV_DISTANCE_TO_EMPTY"
EV_IS_PLUGGED_IN = "EV_IS_PLUGGED_IN"
EV_STATE_OF_CHARGE_MODE = "EV_STATE_OF_CHARGE_MODE"
EV_STATE_OF_CHARGE_PERCENT = "EV_STATE_OF_CHARGE_PERCENT"
EV_TIME_TO_FULLY_CHARGED = "EV_TIME_TO_FULLY_CHARGED"
EXTERNAL_TEMP = "EXT_EXTERNAL_TEMP"
ODOMETER = "ODOMETER"
POSITION_TIMESTAMP = "POSITION_TIMESTAMP"
TIMESTAMP = "TIMESTAMP"
TIRE_PRESSURE_FL = "TYRE_PRESSURE_FRONT_LEFT"
TIRE_PRESSURE_FR = "TYRE_PRESSURE_FRONT_RIGHT"
TIRE_PRESSURE_RL = "TYRE_PRESSURE_REAR_LEFT"
TIRE_PRESSURE_RR = "TYRE_PRESSURE_REAR_RIGHT"
VEHICLE_STATE = "VEHICLE_STATE_TYPE"

# vehicleStatus.json keys
VS_AVG_FUEL_CONSUMPTION = "avgFuelConsumptionLitersPer100Kilometers"
VS_DIST_TO_EMPTY = "distanceToEmptyFuelKilometers"
VS_TIMESTAMP = "eventDate"
VS_LATITUDE = "latitude"
VS_LONGITUDE = "longitude"
VS_HEADING = "positionHeadingDegree"
VS_ODOMETER = "odometerValueKilometers"
VS_VEHICLE_STATE = "vehicleStateType"
VS_TIRE_PRESSURE_FL = "tirePressureFrontLeft"
VS_TIRE_PRESSURE_FR = "tirePressureFrontRight"
VS_TIRE_PRESSURE_RL = "tirePressureRearLeft"
VS_TIRE_PRESSURE_RR = "tirePressureRearRight"


# Erroneous Values
BAD_AVG_FUEL_CONSUMPTION = "16383"
BAD_DISTANCE_TO_EMPTY_FUEL = "16383"
BAD_EV_TIME_TO_FULLY_CHARGED = "65535"
BAD_TIRE_PRESSURE = "32767"
BAD_ODOMETER = None
BAD_SENSOR_VALUES = [
    BAD_AVG_FUEL_CONSUMPTION,
    BAD_DISTANCE_TO_EMPTY_FUEL,
    BAD_EV_TIME_TO_FULLY_CHARGED,
    BAD_TIRE_PRESSURE,
    BAD_ODOMETER,
]

# Timestamp Formats
TIMESTAMP_FMT = "%Y-%m-%dT%H:%M:%S%z"  # "2020-04-25T23:35:55+0000"
POSITION_TIMESTAMP_FMT = "%Y-%m-%dT%H:%M:%SZ"  # "2020-04-25T23:35:55Z"



#  SPDX-License-Identifier: Apache-2.0
"""
subarulink - A Python Package for interacting with Subaru Starlink Remote Services API.

connection.py - provides management for HTTP sessions to Subaru API

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""
import asyncio
import logging
import pprint
import time

import aiohttp
from yarl import URL

from subarulink.exceptions import IncompleteCredentials, SubaruException

_LOGGER = logging.getLogger(__name__)


class Connection:
    """Connection to Subaru Starlink API."""

    def __init__(self, websession: aiohttp.ClientSession, username, password, device_id, device_name,) -> None:
        """Initialize connection object."""
        self.username = username
        self.password = password
        self.device_id = device_id
        self.lock = asyncio.Lock()
        self.device_name = device_name
        self.vehicles = []
        self.vehicle_key = None
        self.default_vin = None
        self.baseurl = "https://mobileapi.prod.subarucs.com/g2v15"
        self.head = {}
        self.head[
            "User-Agent"
        ] = "Mozilla/5.0 (Linux; Android 10; Android SDK built for x86 Build/QSR1.191030.002; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/74.0.3729.185 Mobile Safari/537.36"
        self.head["Origin"] = "file://"
        self.head["X-Requested-With"] = "com.subaru.telematics.app.remote"
        self.head["Accept-Language"] = "en-US,en;q=0.9"
        self.head["Accept-Encoding"] = "gzip, deflate"
        self.head["Accept"] = "*/*"
        self.websession = websession
        self.authenticated = False
        self.registered = False
        self.current_vin = None

    async def connect(self):
        """Connect to and establish session with Subaru Remote Services API."""
        if await self._authenticate():
            await self._refresh_vehicles()
            if self.registered:
                return self.vehicles
            if await self._register_device():
                self.websession.cookie_jar.clear()
                return self.vehicles
        return None

    async def validate_session(self, vin):
        """Validate if current session cookie is still valid with Subaru Remote Services API and vehicle context is correct."""
        result = False
        js_resp = await self.__open("/validateSession.json", "get")
        _LOGGER.debug(pprint.pformat(js_resp))
        if js_resp["success"]:
            if vin != self.current_vin:
                # API call for VIN that is not the current remote context.
                _LOGGER.debug("Switching Subaru API vehicle context to: %s", vin)
                if await self._select_vehicle(vin):
                    result = True
            else:
                result = True
        elif await self._authenticate(vin):
            # New session cookie.  Must call selectVehicle.json before any other API call.
            if await self._select_vehicle(vin):
                result = True
        else:
            self.authenticated = False
        return result

    async def get(self, command, params=None, data=None, json=None):
        """Send HTTPS GET request to Subaru Remote Services API."""
        if self.authenticated:
            return await self.__open(
                f"{command}", method="get", headers=self.head, params=params, data=data, json=json,
            )

    async def post(self, command, params=None, data=None, json=None):
        """Send HTTPS POST request to Subaru Remote Services API."""
        if self.authenticated:
            return await self.__open(
                f"{command}", method="post", headers=self.head, params=params, data=data, json=json,
            )

    async def _authenticate(self, vin=None) -> bool:
        """Authenticate to Subaru Remote Services API."""
        if self.username and self.password and self.device_id:
            post_data = {
                "env": "cloudprod",
                "loginUsername": self.username,
                "password": self.password,
                "deviceId": self.device_id,
                "passwordToken": None,
                "selectedVin": vin,
                "pushToken": None,
                "deviceType": "android",
            }
            js_resp = await self.__open("/login.json", "post", data=post_data, headers=self.head)
            if js_resp.get("success"):
                _LOGGER.debug("Client authentication successful")
                _LOGGER.debug(pprint.pformat(js_resp))
                self.authenticated = True
                self.registered = js_resp["data"]["deviceRegistered"]
                i = js_resp["data"]["currentVehicleIndex"]
                self.current_vin = js_resp["data"]["vehicles"][i]["vin"]
                return True
            elif js_resp.get("errorCode"):
                _LOGGER.debug(pprint.pformat(js_resp))
                error = js_resp.get("errorCode")
                if error == "invalidAccount":
                    _LOGGER.error("Client authentication failed")
                if error == "passwordWarning":
                    _LOGGER.error("Multiple Password Failures.")
                raise SubaruException(error)
            _LOGGER.error("Unknown failure")
            raise SubaruException(js_resp)
        raise IncompleteCredentials("Connection requires email and password and device id.")

    async def _select_vehicle(self, vin):
        """Select active vehicle for accounts with multiple VINs."""
        params = {}
        params["vin"] = vin
        params["_"] = int(time.time())
        js_resp = await self.get("/selectVehicle.json", params=params)
        _LOGGER.debug(pprint.pformat(js_resp))
        if js_resp["success"]:
            self.current_vin = vin
            _LOGGER.debug("Current vehicle: vin=%s", js_resp["data"]["vin"])
            return js_resp["data"]
        self.current_vin = None
        return None

    async def _refresh_vehicles(self):
        js_resp = await self.__open("/refreshVehicles.json", "get", params={"_": int(time.time())})
        _LOGGER.debug(pprint.pformat(js_resp))
        vehicles = js_resp["data"]["vehicles"]
        if len(vehicles) > 1:
            vehicles = await self._refresh_multi_vehicle(vehicles)
        for vehicle in vehicles:
            car = {}
            car["vin"] = vehicle["vin"]
            car["id"] = vehicle["vehicleKey"]
            car["display_name"] = vehicle["vehicleName"]
            if "g2" in vehicle["features"]:
                car["api_gen"] = "g2"
            elif "g1" in vehicle["features"]:
                car["api_gen"] = "g1"
            else:
                car["api_gen"] = "g0"
            if "PHEV" in vehicle["features"]:
                car["hasEV"] = True
            else:
                car["hasEV"] = False
            if "RES" in vehicle["features"]:
                car["hasRES"] = True
            else:
                car["hasRES"] = False
            if "REMOTE" in vehicle["subscriptionFeatures"] and vehicle["subscriptionStatus"] == "ACTIVE":
                car["hasRemote"] = True
            else:
                car["hasRemote"] = False
            if "SAFETY" in vehicle["subscriptionFeatures"] and vehicle["subscriptionStatus"] == "ACTIVE":
                car["hasSafety"] = True
            else:
                car["hasSafety"] = False
            self.vehicles.append(car)

    async def _refresh_multi_vehicle(self, vehicles):
        # refreshVehicles.json returns unreliable data if multiple cars on account
        # use selectVehicle.json to get each car's info
        result = []
        for vehicle in vehicles:
            vin = vehicle["vin"]
            result.append(await self._select_vehicle(vin))
        return result

    async def _register_device(self):
        _LOGGER.debug("Authorizing device via web API")
        web_baseurl = "https://www.mysubaru.com"
        if self.username and self.password and self.device_id:
            post_data = {
                "username": self.username,
                "password": self.password,
                "deviceId": self.device_id,
            }
            resp = await self.__open("/login", "post", data=post_data, baseurl=web_baseurl, decode_json=False)
            if resp:
                js_resp = await self.__open(
                    "/profile/updateDeviceEntry.json", "get", params={"deviceId": self.device_id}, baseurl=web_baseurl,
                )
        if js_resp:
            _LOGGER.info("Device successfully authorized")
            return await self._set_device_name()
        return False

    async def _set_device_name(self):
        _LOGGER.debug("Setting Device Name to %s", self.device_name)
        web_baseurl = "https://www.mysubaru.com"
        js_resp = await self.__open(
            "/profile/addDeviceName.json",
            "get",
            params={"deviceId": self.device_id, "deviceName": self.device_name},
            baseurl=web_baseurl,
        )
        if js_resp:
            _LOGGER.debug("Set Device Name Successful")
            return True
        _LOGGER.debug("Unknown Error during Set Device Name")
        return False

    async def __open(
        self, url, method="get", headers=None, data=None, json=None, params=None, baseurl="", decode_json=True
    ):
        """Open url."""
        if not baseurl:
            baseurl = self.baseurl
        url: URL = URL(baseurl + url)

        _LOGGER.debug("%s: %s", method.upper(), url)
        async with self.lock:
            try:
                resp = await getattr(self.websession, method)(url, headers=headers, params=params, data=data, json=json)
                if resp.status > 299:
                    _LOGGER.debug(pprint.pformat(resp.request_info))
                    _LOGGER.debug(pprint.pformat(resp))
                    raise SubaruException("HTTP %d: %s" % (resp.status, resp))
                if decode_json:
                    return await resp.json()
                return resp
            except aiohttp.ClientResponseError as exception:
                raise SubaruException(exception.status)
            except aiohttp.ClientConnectionError:
                raise SubaruException("aiohttp.ClientConnectionError")



#  SPDX-License-Identifier: Apache-2.0
"""
An example console application that uses the subarulink package.

For more details about this api, please refer to the documentation at
https://github.com/G-Two/subarulink
"""
import argparse
import asyncio
from datetime import datetime
import json
import logging
import os.path
from pprint import pprint
import shelve
import shlex
import sys

from aiohttp import ClientSession
import stdiomask

from subarulink import Controller, SubaruException
import subarulink.const as sc

CONFIG_FILE = ".subarulink.cfg"
LOGGER = logging.getLogger("subarulink")
STREAMHANDLER = logging.StreamHandler()
STREAMHANDLER.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s"))
LOGGER.addHandler(STREAMHANDLER)
LOOP = asyncio.get_event_loop()

OK = "\033[92m"
WARNING = "\033[93m"
FAIL = "\033[91m"
ENDC = "\033[0m"


class CLI:  # pylint: disable=too-few-public-methods
    """A basic shell for interacting with Subaru's Remote Services API."""

    def __init__(self, config_file):
        """Initialize CLI class for subarulink controller."""
        self._config = {}
        self._config_file = config_file
        self._get_config()
        self._ctrl = None
        self._current_vin = None
        self._current_api_gen = None
        self._current_hasEV = None
        self._current_hasRES = None
        self._current_hasRemote = None
        self._session = None
        self._car_data = None
        self._cars = None

    def _get_config(self):
        """Read config file, or create one with user input."""
        saved_config = {}
        write_config = False

        if os.path.isfile(self._config_file):
            LOGGER.info("Opening config file: %s", self._config_file)
            try:
                infile = open(self._config_file, "r")
                config_json = infile.read()
            except UnicodeDecodeError:
                # Update previous version's shelve config file to json
                LOGGER.warning(f"Updating {self._config_file} to JSON format.")
                infile.close()
                self._convert_to_json(self._config_file)
                infile = open(self._config_file, "r")
                config_json = infile.read()
            infile.close()
            saved_config = json.loads(config_json)
        else:
            write_config = True

        self._config = saved_config

        if "username" not in self._config:
            self._config["username"] = input("Enter Subaru Starlink username: ")

        if "password" not in self._config:
            self._config["password"] = stdiomask.getpass("Enter Subaru Starlink password: ")

        if "pin" not in self._config:
            self._config["pin"] = stdiomask.getpass("Enter Subaru Starlink PIN: ")

        if "device_id" not in self._config:
            self._config["device_id"] = int(datetime.now().timestamp())
            write_config = True

        self._config["device_name"] = "subarulink"

        if "save_creds" not in self._config or self._config.get("save_creds") == "N":
            while True:
                save_creds = input("Remember these credentials? [Y]es, [N]o, [D]on't ask again > ")
                if save_creds in ["N", "n"]:
                    break
                if save_creds in ["D", "d"]:
                    saved_config["save_creds"] = save_creds
                    write_config = True
                    break
                if save_creds in ["Y", "y"]:
                    saved_config["save_creds"] = save_creds
                    saved_config["username"] = self._config["username"]
                    saved_config["password"] = self._config["password"]
                    saved_config["pin"] = self._config["pin"]
                    write_config = True
                    break

        if write_config:
            self._save_config()

    def _save_config(self):
        config_to_save = self._config.copy()

        if config_to_save.get("save_creds") not in ["Y", "y"]:
            config_to_save.pop("username")
            config_to_save.pop("password")
            config_to_save.pop("pin")

        with open(self._config_file, "w") as outfile:
            outfile.write(json.dumps(config_to_save))
            LOGGER.info(f"Saved settings to config file: {self._config_file}")
        os.chmod(self._config_file, 0o600)

    def _convert_to_json(self, config_file):
        old_config = {}
        with shelve.open(config_file) as shelf:
            if "username" in shelf:
                old_config["username"] = shelf["username"]
            if "password" in shelf:
                old_config["password"] = shelf["password"]
            if "pin" in shelf:
                old_config["pin"] = shelf["pin"]
            old_config["device_name"] = "subarulink"
            if "device_id" in shelf:
                old_config["device_id"] = shelf["device_id"]
            if "save_creds" in shelf:
                old_config["save_creds"] = shelf["save_creds"]

        os.remove(config_file)
        LOGGER.warning("Deleted %s" % config_file)
        with open(config_file, "w") as outfile:
            outfile.write(json.dumps(old_config))
        LOGGER.info("Saved config file: %s", config_file)
        os.chmod(config_file, 0o600)

    @property
    def _current_name(self):
        return self._ctrl.vin_to_name(self._current_vin)

    async def _quit(self, code):
        await self._session.close()
        sys.exit(code)

    async def _vehicle_select(self, interactive=True, vin=None, reselect=False):
        if (interactive and self._config.get("default_vin") is None) or reselect:
            while True:
                selected = -1
                print("\nAvailable Vehicles:")
                for i in range(len(self._cars)):
                    print("[%d] %s (%s)" % (i + 1, self._ctrl.vin_to_name(self._cars[i]), self._cars[i]))
                if len(self._cars) == 1:
                    selected = 0
                if selected == -1:
                    selected = input("\nSelect Vehicle> ")
                    if not selected.isnumeric():
                        continue
                    selected = int(selected) - 1
                if selected in range(len(self._cars)):
                    break

            self._current_vin = self._cars[selected]

        elif vin:
            if vin in self._cars:
                self._current_vin = vin
            else:
                LOGGER.error(f"VIN {vin} does not exist in user account.")
                await self._quit(3)

        elif len(self._cars) == 1:
            self._current_vin = self._cars[0]

        elif self._config.get("default_vin") in self._cars:
            self._current_vin = self._config.get("default_vin")

        else:
            LOGGER.error("Multiple vehicles in account but VIN not specified in config or command line")
            await self._quit(1)

        self._current_hasEV = self._ctrl.get_ev_status(self._current_vin)
        self._current_hasRES = self._ctrl.get_res_status(self._current_vin)
        self._current_hasRemote = self._ctrl.get_remote_status(self._current_vin)
        self._current_api_gen = self._ctrl.get_api_gen(self._current_vin)

    async def _set_climate_params(self):
        modes = [
            sc.MODE_AUTO,
            sc.MODE_FACE,
            sc.MODE_FEET,
            sc.MODE_SPLIT,
            sc.MODE_FEET_DEFROST,
            sc.MODE_DEFROST,
        ]
        speeds = [
            sc.FAN_SPEED_AUTO,
            sc.FAN_SPEED_LOW,
            sc.FAN_SPEED_MED,
            sc.FAN_SPEED_HI,
        ]
        seat_heat = [sc.HEAT_SEAT_OFF, sc.HEAT_SEAT_LOW, sc.HEAT_SEAT_MED, sc.HEAT_SEAT_HI]
        defrost = [sc.REAR_DEFROST_OFF, sc.REAR_DEFROST_ON]
        recirculate = [sc.RECIRCULATE_OFF, sc.RECIRCULATE_ON]
        rear_ac = [sc.REAR_AC_OFF, sc.REAR_AC_ON]
        self._config["hvac"] = {}
        while True:
            print("Enter temperature (%d-%d):" % (sc.TEMP_MIN, sc.TEMP_MAX))
            hvac_temp = input("> ")
            if hvac_temp.isnumeric():
                if sc.TEMP_MIN < int(hvac_temp) < sc.TEMP_MAX:
                    break

        self._config["climate"] = {}
        self._config["climate"][sc.TEMP] = hvac_temp
        self._config["climate"][sc.MODE] = _select_from_list("Select mode:", modes)
        self._config["climate"][sc.FAN_SPEED] = _select_from_list("Select fan speed:", speeds)
        self._config["climate"][sc.HEAT_SEAT_LEFT] = _select_from_list("Driver seat heat:", seat_heat)
        self._config["climate"][sc.HEAT_SEAT_RIGHT] = _select_from_list("Passenger seat heat:", seat_heat)
        self._config["climate"][sc.REAR_DEFROST] = _select_from_list("Rear defroster:", defrost)
        self._config["climate"][sc.RECIRCULATE] = _select_from_list("Recirculate:", recirculate)
        self._config["climate"][sc.REAR_AC] = _select_from_list("Rear AC:", rear_ac)
        save = _select_from_list("Save HVAC settings?", ["Yes", "No"])
        if save == "Yes":
            pprint(self._config["climate"])
            await self._ctrl.save_climate_settings(self._current_vin, self._config["climate"])

    async def _fetch_climate_settings(self):
        success = await self._ctrl.get_climate_settings(self._current_vin)
        if success:
            await self._fetch()
            pprint(self._car_data["climate"])

    async def _remote_start(self, args):
        if len(args) == 0:
            print("\nremote_start [set|show|start|stop]")
            print("  set   - enter climate settings")
            print("  show  - show saved climate settings")
            print("  on    - start engine")
            print("  off   - stop engine\n")
        elif args[0] == "set":
            await self._set_climate_params()
        elif args[0] == "show":
            await self._fetch_climate_settings()
        elif args[0] == "off":
            await self._ctrl.remote_stop(self._current_vin)
        elif args[0] == "on":
            if self._car_data.get("climate") is None:
                await self._fetch_climate_settings()
            await self._ctrl.remote_start(self._current_vin, self._car_data["climate"])
        else:
            print("remote_start: invalid arg: %s" % args[0])

    def _show(self, args):
        if len(args) != 1:
            print("\nshow [summary|all]")
            print("  summary - display summary information")
            print("  all     - display all data available\n")
        elif args[0] == "all":
            pprint(self._car_data)
        elif args[0] == "summary":
            timediff = datetime.now() - datetime.fromtimestamp(self._car_data["status"][sc.TIMESTAMP])
            print(
                "\nVehicle last reported data %d days, %d hours, %d minutes ago \n"
                % (timediff.days, timediff.seconds // 3600, (timediff.seconds) // 60 % 60,)
            )
            # Safety Plus Data
            print("Odometer: %0.1f miles" % _km_to_miles(self._car_data["status"][sc.ODOMETER]))
            print("Distance to Empty: %d miles" % _km_to_miles(self._car_data["status"][sc.DIST_TO_EMPTY]))
            print(
                "Average Fuel Consumption: %d MPG" % _L100km_to_mpg(self._car_data["status"][sc.AVG_FUEL_CONSUMPTION])
            )
            # Lat/Long assumes North America hemispheres since Starlink is a Subaru of America thing
            if self._car_data["status"].get(sc.LATITUDE) and self._car_data["status"].get(sc.LONGITUDE):
                print(
                    "Position: %f°N  %f°W  Heading: %d"
                    % (
                        self._car_data["status"].get(sc.LATITUDE),
                        (self._car_data["status"].get(sc.LONGITUDE) or 0) * -1,
                        (self._car_data["status"].get(sc.HEADING) or 0),
                    )
                )
            print("Vehicle State: %s" % self._car_data["status"][sc.VEHICLE_STATE])
            print("Tire Pressures (psi):")
            print(
                "  FL: %d   FR: %d "
                % (
                    _kpa_to_psi(self._car_data["status"][sc.TIRE_PRESSURE_FL]),
                    _kpa_to_psi(self._car_data["status"][sc.TIRE_PRESSURE_FR]),
                )
            )
            print(
                "  RL: %d   RR: %d "
                % (
                    _kpa_to_psi(self._car_data["status"][sc.TIRE_PRESSURE_RL]),
                    _kpa_to_psi(self._car_data["status"][sc.TIRE_PRESSURE_RR]),
                )
            )

            # Security Plus Data
            if self._current_hasRemote:
                print("External Temp: %0.1f °F" % _c_to_f(self._car_data["status"][sc.EXTERNAL_TEMP]))

            # EV Data
            if self._current_hasEV:
                print("EV Charge: %s%%" % self._car_data["status"][sc.EV_STATE_OF_CHARGE_PERCENT])
                print("Aux Battery: %sV" % self._car_data["status"][sc.BATTERY_VOLTAGE])
                print("EV Plug Status: %s" % self._car_data["status"][sc.EV_IS_PLUGGED_IN])
                print("EV Distance to Empty: %s miles" % self._car_data["status"][sc.EV_DISTANCE_TO_EMPTY])
        else:
            print("show: invalid arg: %s" % args[0])

    async def _fetch(self):
        LOGGER.info("Fetching data for %s..." % self._ctrl.vin_to_name(self._current_vin))
        self._car_data = await self._ctrl.get_data(self._current_vin)
        return True

    async def _connect(self, interactive=True, vin=None):
        LOGGER.info("Connecting to Subaru Remote Services API")
        try:
            if await self._ctrl.connect():
                LOGGER.info("Successfully connected")
                self._cars = self._ctrl.get_vehicles()
                await self._vehicle_select(interactive, vin)
                if interactive and self._current_api_gen == "g2":
                    await self._fetch()
                    self._show(["summary"])
                elif self._current_api_gen == "g1":
                    LOGGER.warning(
                        "%s is a Generation 1 telematics vehicle which has not been tested." % self._current_name
                    )
                elif not interactive:
                    pass
                else:
                    LOGGER.error("Unknown telematics version: %s" % self._current_api_gen)
        except SubaruException:
            LOGGER.error("Unable to connect.  Check Username/Password.")
            await self._session.close()
            return False
        return True

    async def _cli_loop(self):
        print("\nEnter a command. For a list of commands, enter '?'.")

        while True:
            print("%s" % self._current_name, end="")
            try:
                cmd, *args = shlex.split(input("> "))
            except ValueError:
                continue

            try:
                if cmd == "quit":
                    await self._quit(0)

                elif cmd in ["help", "?"]:
                    print("\nCommands:")
                    print("  help         - display this help")
                    if len(self._cars) > 1:
                        print("  vehicle      - change vehicle")
                        print("  default      - set this vehicle as default")
                    print("  lock         - lock vehicle doors")
                    print("  unlock       - unlock vehicle doors")
                    print("  lights       - turn on lights")
                    print("  horn         - sound horn")
                    if self._current_api_gen == "g2":
                        print("  show         - show vehicle information")
                        print("  update       - request update from vehicle")
                        print("  fetch        - fetch most recent update")
                    if self._current_hasEV:
                        print("  charge       - start EV charging")
                    if self._current_hasRES or self._current_hasEV:
                        print("  remote_start - remote start")
                    print("  quit\n")

                elif cmd == "vehicle":
                    await self._vehicle_select(reselect=True)

                elif cmd == "default":
                    self._config["default_vin"] = self._current_vin
                    self._save_config()

                elif cmd == "lock":
                    await self._ctrl.lock(self._current_vin)

                elif cmd == "unlock":
                    await self._ctrl.unlock(self._current_vin)

                elif cmd == "lights":
                    await self._ctrl.lights(self._current_vin)

                elif cmd == "horn":
                    await self._ctrl.horn(self._current_vin)

                elif cmd == "show" and self._current_api_gen == "g2":
                    self._show(args)

                elif cmd == "update" and self._current_api_gen == "g2":
                    await self._ctrl.update(self._current_vin)
                    await self._fetch()

                elif cmd == "fetch" and self._current_api_gen == "g2":
                    await self._fetch()

                elif cmd == "charge" and self._current_hasEV:
                    await self._ctrl.charge_start(self._current_vin)

                elif cmd == "remote_start" and (self._current_hasRES or self._current_hasEV):
                    await self._remote_start(args)

                else:
                    print("invalid command: {}".format(cmd))

            except SubaruException as exc:
                LOGGER.error("SubaruException caught: %s", exc.message)

    async def run(self):
        """Initialize connection and start CLI loop."""
        self._session = ClientSession()
        self._ctrl = Controller(
            self._session,
            self._config["username"],
            self._config["password"],
            self._config["device_id"],
            self._config["pin"],
            self._config["device_name"],
        )
        try:
            if await self._connect():
                await self._cli_loop()
        except (KeyboardInterrupt, EOFError):
            await self._quit(0)

    async def single_command(self, cmd, vin):
        """Initialize connection and execute as single command."""
        success = False
        self._session = ClientSession()
        self._ctrl = Controller(
            self._session,
            self._config["username"],
            self._config["password"],
            self._config["device_id"],
            self._config["pin"],
            self._config["device_name"],
        )

        if await self._connect(interactive=False, vin=vin):
            try:
                if cmd == "status":
                    success = await self._fetch()
                    pprint(self._car_data)

                elif cmd == "lock":
                    success = await self._ctrl.lock(self._current_vin)

                elif cmd == "unlock":
                    success = await self._ctrl.unlock(self._current_vin)

                elif cmd == "lights":
                    success = await self._ctrl.lights(self._current_vin)

                elif cmd == "horn":
                    success = await self._ctrl.horn(self._current_vin)

                elif cmd == "locate":
                    success = await self._ctrl.update(self._current_vin)
                    await self._fetch()
                    pprint(self._car_data.get("location"))

                elif cmd == "remote_start":
                    success = await self._ctrl.remote_start(self._current_vin)

                elif cmd == "remote_stop":
                    success = await self._ctrl.remote_stop(self._current_vin)

                elif cmd == "charge":
                    success = await self._ctrl.charge_start(self._current_vin)

                else:
                    LOGGER.error("Unsupported command")

            except SubaruException as exc:
                LOGGER.error("SubaruException caught: %s", exc.message)

        if success:
            print(f"{OK}Command '{cmd}' completed for {self._current_vin}{ENDC}")
            sys.exit(0)
        else:
            print(f"{FAIL}Command '{cmd}' failed for {self._current_vin}{ENDC}")
            sys.exit(1)


def _km_to_miles(meters):
    return float(meters) * 0.62137119


def _c_to_f(temp_c):
    return float(temp_c) * 1.8 + 32.0


def _L100km_to_mpg(L100km):
    return round(235.215 / L100km, 1)


def _kpa_to_psi(kpa):
    return kpa / 68.95


def _select_from_list(msg, items):
    while True:
        print(msg)
        for i, val in enumerate(items):
            print(" [%d] %s" % (i + 1, val))
        choice = input("> ")
        if choice.isnumeric():
            choice = int(choice) - 1
            if choice in range(len(items)):
                return items[choice]


def main():
    """Run a basic CLI that uses the subarulink package."""
    home_dir = os.path.expanduser("~")
    default_config = os.path.join(home_dir, CONFIG_FILE)

    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers(title="command", description="execute single command and exit", dest="command")
    status_command = subparsers.add_parser("status", help="get vehicle status information")
    status_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    lock_command = subparsers.add_parser("lock", help="lock doors")
    lock_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    unlock_command = subparsers.add_parser("unlock", help="unlock doors")
    unlock_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    lights_command = subparsers.add_parser("lights", help="turn on lights")
    lights_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    horn_command = subparsers.add_parser("horn", help="sound horn")
    horn_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    locate_command = subparsers.add_parser("locate", help="locate vehicle")
    locate_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    start_command = subparsers.add_parser("remote_start", help="remote engine start")
    start_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    stop_command = subparsers.add_parser("remote_stop", help="remote engine stop")
    stop_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")
    charge_command = subparsers.add_parser("charge", help="start PHEV charging")
    charge_command.add_argument("--vin", required=False, help="VIN (required if not specified in config file)")

    parser.add_argument("-i", "--interactive", help="interactive mode", action="store_true", dest="interactive")
    parser.add_argument(
        "-c",
        "--config",
        default=default_config,
        help=f"specify config file (default is {default_config}",
        dest="config_file",
    )
    parser.add_argument(
        "-v",
        "--verbosity",
        type=int,
        choices=[0, 1, 2],
        default=0,
        help="verbosity level: 0=error[default] 1=info 2=debug",
    )

    args = parser.parse_args()

    if args.verbosity == 1:
        LOGGER.setLevel(logging.INFO)
    elif args.verbosity == 2:
        LOGGER.setLevel(logging.DEBUG)
    else:
        LOGGER.setLevel(logging.ERROR)

    if args.interactive and args.command:
        print("Error - Cannot select both interactive mode and single command mode")
        sys.exit(4)

    if args.command:
        if not os.path.isfile(args.config_file):
            LOGGER.error(
                f"Config file '{args.config_file}' not found.  Please run interactively once before using single command mode."
            )
            sys.exit(2)
        LOGGER.info(f"Entering Single command mode: cmd={args.command}, vin={args.vin}")
        cli = CLI(args.config_file)
        LOOP.run_until_complete(cli.single_command(args.command, args.vin))
    if args.interactive:
        LOGGER.info("Entering interactive mode")
        cli = CLI(args.config_file)
        LOOP.run_until_complete(cli.run())
    else:
        parser.print_help()
